require_relative 'sr/core_patches'
require_relative 'sr/version'
require 'sxdg/as_xdg'

# `sr` is a simple, tiny, hackable REPL for Ruby.
module SR
  class << self
    # @return [Binding] the currently bound SR context
    attr_reader :context

    # @return [Hash{Class=>#call}] the hash of exception handlers
    attr_reader :exception_handlers

    # @return [#call] the input handler
    attr_accessor :input_handler

    # @return [Hash{Symbol=>#call,#to_s}] the hash of prompts to display; valid
    #   keys are `:error`, `:input`, `:more`, and `:return`; valid values must
    #   respond to `#to_s`
    attr_reader :prompt
  end

  # The array of known configuration files for SR which should be loaded when
  # the read-eval-print loop is started.
  CONFIGURATION_FILES = (XDG_CONFIG_DIRS + [XDG_CONFIG_HOME]).map! do |dir|
    (File.expand_path(dir) + '/sr/config.rb').freeze
  end.select! { |file| File.exist?(file) }.freeze

  # The regular expression used to determine whether or not a syntax error is
  # a valid continuation which should prompt for more user input.
  CONTINUATION_REGEXP =
    /unterminated (?:regexp|string)|unexpected end-of-(?:file|input)|tIDENTIFIER/

  @prompt = {
    error:  proc { |e| "!! #{e.class}: #{e.message}" },
    input:  proc do
      c = @context
      "== #{c.respond_to?(:receiver) ? c.receiver : c.eval('self')}:\n>> "
    end,
    more:   '..   ',
    return: proc { |v| "-> #{v.inspect}" },
  }

  @input_handler ||= ->(more) do
    print(call_prompt(more ? :more : :input))
    input = STDIN.gets
    input = "#{more}#{input}" if more
    input
  end

  @exception_handlers = {
    SyntaxError => proc do |e, i|
      e.message =~ CONTINUATION_REGEXP ? read_eval_print(i) : report(e)
    end,
    Interrupt => proc do |_, _, m|
      m ? (puts ; read_eval_print) : @enabled = false
    end,
    SystemExit => proc { @enabled = false },
    Exception => proc { |e| report(e) },
  }

  @context = @ORIGINAL_CONTEXT = TOPLEVEL_BINDING

  @stack = []
  
  # @return [Array<Binding>] a copy of the SR context stack
  def self.stack
    @stack.dup
  end

  # @param object [#__binding__] the object to bind the SR context to
  # @return [Binding] the new SR context
  def self.bind(object)
    @stack.push(@context)
    @context = object.__binding__
  end
  class << self ; alias_method :context=, :bind ; end

  # @return [Binding] the previous binding in the SR context stack or the
  #   default context if no previous binding exists
  def self.unbind
    @context = @stack.empty? ? @ORIGINAL_CONTEXT : @stack.pop
  end

  # @param context [#__binding__] the context to use as the new default
  #   context; defaults to the top level binding if no context given
  # @return [Binding] the reset binding context
  def self.reset_binding(context = nil)
    @stack.clear
    context  = context.__binding__ unless context.nil?
    @context = @ORIGINAL_CONTEXT = context ? context : TOPLEVEL_BINDING
  end

  # @param key [Symbol] the key of the @prompt hash to call or fetch
  # @param args [Array<Object>] the arguments to be passed to the value of the
  #   given key if the value is a proc
  # @return [String] the prompt to display
  private_class_method def self.call_prompt(key, *args)
    prompt = @prompt[key]
    prompt.respond_to?(:call) ? prompt.call(*args) : (prompt || '?? ')
  end

  # @param ex [Exception] the exception to report
  # @return [void]
  private_class_method def self.report(ex)
    @context.local_variable_set(:_, ex)
    $stderr.puts call_prompt(:error, ex) if @prompt[:error]
  end

  # @param more [String] the previous input to be evaluated along with the
  #   given input
  # @return [Object] the return value of the given input
  private_class_method def self.read_eval_print(more = nil)
    value = @context.eval(input = @input_handler.call(more), '(sr)', 0)
    @context.local_variable_set(:_, value)
    puts call_prompt(:return, value) if @prompt[:return]
  rescue *@exception_handlers.keys => ex
    @exception_handlers.find { |(c,_)| ex.is_a?(c) }[1].call(ex, input, more)
  end

  # Loads the configuration for SR, resets the default binding to the given
  # `context`, and starts the read-eval-print loop.
  #
  # @param context [#__binding__] the object providing the default context for
  #   `sr`; defaults to the top level binding if no context given
  # @param options [Hash] the options for the REPL; valid options are
  #   `:config_files` (`Array<String>`) and `:context` (`#__binding__`)
  # @return [void]
  def self.repl(context = nil, **options)
    options.fetch(:config_files, CONFIGURATION_FILES).each(&method(:load))
    context = context || options[:context]
    reset_binding(context) unless context.nil?
    @enabled = true
    read_eval_print while @enabled
  end
end
